
(* persistent union-find = Tarjan's algorithm with persistent arrays *)

module Make(A : Sig.PersistentArray) : Sig.UnionFind = struct

  type t = { 
    c: A.t;
    mutable father: A.t;
  }
      
  let create n = 
    { c = A.create n 0;
      father = A.init n (fun i -> i) }

  let rec find_aux f i = 
    let fi = A.get f i in
    if fi == i then 
      f, i
    else 
      let f, r = find_aux f fi in 
      let f = A.set f i r in
      f, r
      
  let find h x = 
    let f,rx = find_aux h.father x in h.father <- f; rx
  
  let union h x y = 
    let rx = find h x in
    let ry = find h y in
    if rx != ry then begin
      let rxc = A.get h.c rx in
      let ryc = A.get h.c ry in
      if rxc > ryc then
	{ h with father = A.set h.father ry rx }
      else if rxc < ryc then
	{ h with father = A.set h.father rx ry }
      else
	{ c = A.set h.c rx (rxc + 1);
	  father = A.set h.father ry rx }
    end else
      h

end
